const cluster = require('cluster');
const dgram = require('dgram');
const hash = require('./hash');
// const os = require('os').cpus().length;

// Creates and handles workers to find solutions to problems.
if (cluster.isMaster) {
  const channels = {};
  const client = dgram.createSocket('udp4');

  // Returns the product of the factory function.
  client.connect(8888, 'nate.ml', () => {
    client.on('error', (e) => {
      console.log(e);
    });
  
    client.on('message', (msg) => {
      // Message expected to be valid JSON formatting.
      let message = msg.toString();
  
      // If the message is empty, it is likely a heartbeat.
      if (message === '') client.send(Buffer.from(''), 8888, 'nate.ml');
  
      // Drops the connection if the message is not valid JSON.
      try { message = JSON.parse(message); } catch (e) { return; }
  
      // Checks to see if there is a valid channel for the message.
      if (
        typeof message.channel === 'string'
        && typeof message.data === 'object'
        && typeof channels[message.channel] === 'function'
      ) channels[message.channel](message.data);
    });

    const listen = (id, callback) => {
      channels[id] = callback;
      return;
    };

    const send = (channel, data) => {
      const response = Buffer.from(JSON.stringify({ channel, data }));
      client.send(response, 8888, 'nate.ml');
      return;
    };

    client.send(Buffer.from(''), 8888, 'nate.ml');

    function updateDisplay() {
      console.clear();
      console.log(`Toshizuno's Kryptoin Currency (mining as ${account.username})`);
      console.log('');
      console.log(`Balance: ${balance} KPT`);
      console.log(`Success Rate: ${successfulSolutions}/${totalSolutions}`);
      console.log(`Assigned CPUs: ${cpus}`);
      let ams = 0;
      miningSpeedHistory.forEach((value) => {
        ams += value;
      });
  
      ams = ams / miningSpeedHistory.length;
      if (miningSpeedHistory.length === 0) ams = 0;
      console.log(`Average Mining Speed: ${Math.floor(ams)}ms`);
    }
  
    const account = {
      username: `nathan${Math.floor(Math.random() * 5)}`,
      password: 'password',
    };
  
    let algorithm = null;
    let difficulty = null;
    let problem = null;
    let balance = 0;
    let totalSolutions = -1;
    let successfulSolutions = 0;
    let miningSpeedHistory = [];
  
    const cpus = Math.floor(Math.random() * 4) + 1;
    let workers = [];
  
    listen('update', (data) => {
      algorithm = data.algorithm;
      difficulty = data.difficulty;
      problem = data.problem;
      winner = data.winner;
      if (winner) successfulSolutions += 1;
      totalSolutions += 1;
      updateDisplay();
  
      // Deletes any existing workers.
      for (let i = 0; i < workers.length; i += 1) {
        workers[i].process.kill();
      }
  
      workers = [];
  
      const time = new Date().getTime();
  
      // Creates new workers.
      for (let i = 0; i < cpus; i += 1) {
        workers.push(cluster.fork({
          algorithm,
          cpus,
          difficulty,
          problem,
          startingNumber: i,
        }).on('message', (message) => {
          if (miningSpeedHistory.length >= 50) miningSpeedHistory.pop();
          miningSpeedHistory.push(new Date().getTime() - time);
          // console.log(`> ${account.username}:${account.password}:${message}`);
          send('solve', { solution: message });
  
          // console.log(`\x1b[33m${message}\x1b[0m`);
          // console.log(`took ${new Date().getTime() - time}ms`);
          for (let i = 0; i < workers.length; i += 1) {
            workers[i].process.kill();
          }
  
          workers = [];
        }));
      }
    });
  
    send('query', {});
  });
}

// Creates a worker that can find a solution to a problem.
if (!cluster.isMaster) {
  const algorithm = process.env['algorithm'];
  const cpus = Number(process.env['cpus']);
  const difficulty = Number(process.env['difficulty']);
  const problem = process.env['problem'];
  let currentNumber = Number(process.env['startingNumber']);

  // Attempts to find a solution to the problem.
  new Promise((resolve) => {
    do {
      const solution = String(currentNumber);
      const hashedSolution = hash(algorithm, `${problem}${solution}`);
      // DEBUG/SHOWOFF CODE: console.log(hash);
      if (hashedSolution.endsWith('0'.repeat(difficulty))) return resolve(solution);
      currentNumber += cpus;
    } while (true);
  }).then((solution) => {
    process.send(solution);
  });
}
