const crypto = require('crypto');
const dgram = require('dgram');

const accounts = {
  nathanalex: {
    balance: 0,
    password: 'password',
  }
};

const clients = {};

let algorithm = 'sha256';
let difficulty = 8;
let problem = crypto.createHash(algorithm).update(`${new Date().getTime()}`).digest('hex');

function display() {
  console.clear();
  console.log(JSON.stringify(accounts, null, 2));
  console.log('');
  console.log(JSON.stringify(clients, null, 2));
}

setInterval(function checkHeartbeats() {
  const currentTime = new Date().getTime();
  Object.entries(clients).forEach((client) => {
    console.log(clients[client[0]]);
    if (currentTime - client[1].timeout > 120000) {
      delete clients[client[0]]; // 120000
    }
  });

  display();
}, 1000);

const server = dgram.createSocket('udp4', (msg, rinfo) => {
  // Message expected to be valid JSON formatting.
  let message = msg.toString();

  // Drops the connection if the message is not valid JSON.
  try { message = JSON.parse(message); } catch (e) { return; }

  // Checks to see if there is a valid channel for the message.
  if (
    typeof message.channel === 'string'
    && typeof message.data === 'object'
  ) {
    // heartbeat (keep-alive) 2min
    if (message.channel === 'heartbeat') {
      clients[`${rinfo.address}:${rinfo.port}`].timeout = new Date().getTime();
      clients[`${rinfo.address}:${rinfo.port}`].heartbeat = setInterval();
      display();
      return;
    }

    // sends current balance
    if (
      typeof message.account === 'object'
      && typeof message.account.username === 'string'
      && typeof message.account.password === 'string'
      && typeof accounts[message.account.username] === 'object'
      && message.account.password === accounts[message.account.username].password
      && message.channel === 'requestBalance'
    ) {
      const channel = '';
      const data = {
        balance: accounts[message.account.username].balance,
      };

      //
      const response = Buffer.from(JSON.stringify({ channel, data }));
      server.send(response, rinfo.port, rinfo.address);
      return;
    }

    // sends current problem
    if (message.channel === 'requestProblem') {
      const channel = '';
      const data = {
        algorithm,
        difficulty,
        problem,
      };

      //
      const response = Buffer.from(JSON.stringify({ channel, data }));
      server.send(response, rinfo.port, rinfo.address);
      return;
    }

    // solves current problem
    if (message.channel === 'solveProblem') {
      const solution = crypto.createHash(algorithm).update(problem + message.solution).digest('hex');

      // The solution was correct.
      if (solution.endsWith('0'.repeat(difficulty))) {
        //
        return;
      }

      // The solution was not correct.
      return;
    }
  }
});

server.on('error', (err) => {
  console.log(`server error:\n${err.stack}`);
  server.close();
});

server.on('listening', () => {
  const address = server.address();
  console.log(`server listening ${address.address}:${address.port}`);
});

server.bind(1025);
