const dgram = require('dgram');
const problem = require('./problem')('sha256', 4);

const channels = {};
const clients = {};
const server = dgram.createSocket('udp6');

// Removes any clients that have not been active for two minutes.
setInterval(() => {
  Object.entries(clients).forEach(([key, value]) => {
    const adr = key.split(':');
    clients[key] += 1;
    if (value >= 2) delete clients[key];
    server.send(Buffer.from(''), adr[1], adr[0]);
  });
}, 60000);

// Handles messages sent from clients to the server.
server.on('message', (msg, rinfo) => {
  // Message expected to be valid JSON formatting.
  let message = msg.toString();

  // If the message is empty, it is likely a heartbeat.
  if (message === '') clients[`${rinfo.address}:${rinfo.port}`] = 0;

  // Drops the connection if the message is not valid JSON.
  try { message = JSON.parse(message); } catch (e) { return; }

  // Checks to see if there is a valid channel for the message.
  if (
    typeof message.channel === 'string'
    && typeof message.data === 'object'
    && typeof channels[message.channel] === 'function'
  ) channels[message.channel](message.data, { address: rinfo.address, port: rinfo.port });
});

// Makes the server accessible from the internet.
server.bind(8888, '::');

const formattedClients = () => {
  const temp = [];
  Object.keys(clients).forEach((client) => {
    const adr = client.split(':');
    temp.push({
      address: adr[0],
      port: Number(adr[1]),
    });
  });

  return temp;
}

const listen = (id, callback) => {
  channels[id] = callback;
  return;
};

const send = ({ address, channel, data, port }) => {
  const response = Buffer.from(JSON.stringify({ channel, data }));
  server.send(response, port, address);
  return;
};

listen('solve', (data, rinfo) => {
  if (problem.solve(data.solution)) {
    formattedClients().forEach((client) => {
      send({
        address: client.address,
        channel: 'update',
        data: {
          algorithm: problem.get().algorithm,
          difficulty: problem.get().difficulty,
          problem: problem.get().problem,
          winner: `${client.address}:${client.port}` === `${rinfo.address}:${rinfo.port}` ? true : false,
        },
        port: client.port,
      });
    });
  } else {
    send({
      address: rinfo.address,
      channel: 'update',
      data: {
        algorithm: problem.get().algorithm,
        difficulty: problem.get().difficulty,
        problem: problem.get().problem,
        winner: false,
      },
      port: rinfo.port,
    });
  }
});

listen('query', (data, rinfo) => {
  console.log(rinfo);
  send({
    address: rinfo.address,
    channel: 'update',
    data: {
      algorithm: problem.get().algorithm,
      difficulty: problem.get().difficulty,
      problem: problem.get().problem,
      winner: false,
    },
    port: rinfo.port,
  });
});

setInterval(() => {
  console.clear();
  console.log(problem.get());
  console.table(formattedClients());
}, 1000);
